// file: $ENDPOINT/api/vad.cc
//

// system include files
//
#include <string.h>
#include <unistd.h>

// isip include files
//
#include "../integral_constants.h"
#include "../signal_detector.h"
#include "../signal_detector_constants.h"

// define a format for outputing endpoint information
//
#define SIGD_OUTPUT_FORMAT "%4ld beg_time = %9.3f secs   end_time = %9.3f secs\n"

// define the grow size of the endpoints buffer
//
#define SIGD_ENDPTS_GROW_SIZE 10000
#define NUM_QUIT_TRY 100

// local functions
//
extern int_4 grow_arrays_cc(float_4** arrays, int_4 num_arrays,
			    int_4& cur_size, int_4 grow_size);

//-----------------------------------------------------------------------------
//
// program: signal_detector
// synopsis: signal_detector.exe [options] files
// descr: automatically locates utterances in a file
// example: signal_detector file.param signal.raw signal.endpoints
//
// options: none
//
// arguments:
//  file.param: (input) parameter file
//  signal.raw: (input) binary signal file
//  signal.endpoints: (output) contains beg and end times of utterances
//
//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
  
  // declare an output buffer
  //
  int_4 endpts_max_size = SIGD_ENDPTS_GROW_SIZE;
  float_4* endpts[2];
  endpts[(int_4)0] = new float_4[endpts_max_size];
  endpts[(int_4)1] = new float_4[endpts_max_size];
  
  // declare signal-related objects
  //
  Signal_detector sig_det;
  
  // check if all the required file names have been specified by the user
  //
  if(argc != 4){
    printf("%s\n","usage: vad.exe <parameter file> <input raw file> <output raw file>");
    return ISIP_FALSE;
  }
  
  // load parameters
  //
  if (sig_det.load_parameters_cc((char_1*)argv[1]) == ISIP_FALSE) {
    fprintf(stdout, "%s: error reading parameters (%s)\n", argv[0], argv[1]);
    return(ISIP_ERROR);
  }
  sig_det.debug_cc(stdout, (char_1*)"signal_detector_main");
  
  
  // open the input file
  //
  FILE* fp_in = fopen(argv[2], "r");
  while (fp_in == NULL) {
    sleep (1);
    fp_in = fopen(argv[2], "r");
  }
    
  // open the output file
  //
  FILE* fp_out = fopen(argv[3], "w");
  
  if (fp_out == NULL) {
    fprintf(stdout, "%s: error opening output file (%s)\n", argv[0], argv[3]);
    return(ISIP_ERROR);
  }

  // get the amount of data to read per i/o call
  //
  int_4 nbytes_in = sig_det.get_nbytes_to_read_cc();
  fprintf(stdout, "input buffer size = %ld bytes\n", nbytes_in);

  uint_1* signal_buf = new uint_1[nbytes_in];
  memset(signal_buf, (int_4)0, nbytes_in);

  // initialize the signal detector
  //
  if (sig_det.init_cc() == ISIP_FALSE) {
    fprintf(stdout, "%s: error during initialization\n", argv[0]);
    return(ISIP_ERROR);
  }

  //---------------------------------------------------------------------------
  //
  // main processing loop
  //
  //---------------------------------------------------------------------------

  // loop over the data
  //
  int_4 nf = (int_4)0;
  int_4 num_utt = (int_4)0;
  float_4 t1, t2;
  int_4 num_quit_try = NUM_QUIT_TRY;
  int_4 quit_utt = 0;
  int frame = 0;
  
  // check to see if an utterance has been detected. quit reading the data
  // only if a valid utterance has been detected.
  //
  while (1) {

    // read a buffer full of data
    //
    int num_read = fread(signal_buf,(int_4)1, nbytes_in, fp_in);
    frame++;
    
    // No more data available: after 'quit_check' times of checking,
    // then quit
    //
    if (num_read == 0) {
      quit_utt = (num_quit_try == NUM_QUIT_TRY) ? num_utt : quit_utt;
      num_quit_try--;
      if (quit_utt != num_utt) {
	num_quit_try = NUM_QUIT_TRY;
      }
      else if ((num_quit_try == 0) && (quit_utt == num_utt)) {
	break;
      }
    }

    // process data
    //
    if (sig_det.process_cc((void*)signal_buf) == ISIP_FALSE) {
      fprintf(stdout, "%s: error processing frame no. %ld\n", argv[0], nf);
      return(ISIP_ERROR);
    }
    
    // check for an utterance
    //
    if (sig_det.get_endpoints_cc(t1, t2) == ISIP_TRUE) {

      // make sure we have space for the new endpts
      //
      if (num_utt == endpts_max_size) {
	grow_arrays_cc(endpts, (int_4)2,
		       endpts_max_size, SIGD_ENDPTS_GROW_SIZE);
      }

      // save the endpoints
      //
      (endpts[(int_4)0])[num_utt] = t1;
      (endpts[(int_4)1])[num_utt] = t2;

      // send some output to the display and the output file
      //
      fprintf(stdout, SIGD_OUTPUT_FORMAT, num_utt,
	      (endpts[(int_4)0])[num_utt], (endpts[(int_4)1])[num_utt]);
      
      // increment utterance count if an utterance has been detected
      //
      num_utt++;
      sig_det.reset_cc();
    }
    
    // increment frame counters and zero out the input buffer
    //
    memset(signal_buf, (int_4)0, nbytes_in);
    nf++;
    
    //-------------------------------------------------------------------------
    //
    // cleanup
    //
    //-------------------------------------------------------------------------

    // check one last time for an utterance
    //
    if (sig_det.get_endpoints_cc(t1, t2) == ISIP_TRUE) {
      
      // make sure we have space for the new endpts
      //
      if (num_utt == endpts_max_size) {
	grow_arrays_cc(endpts, (int_4)2,
		       endpts_max_size, SIGD_ENDPTS_GROW_SIZE);
      }
      
      // save the endpoints
      //
      (endpts[(int_4)0])[num_utt] = t1;
      (endpts[(int_4)1])[num_utt] = t2;
      
      // increment utterance count
      //
      num_utt++;
      sig_det.reset_cc();  
    }
  }

  //---------------------------------------------------------------------------
  //
  // excise the endpointed part of the input signal 
  //
  //---------------------------------------------------------------------------

  // get the parameters from the algorithm chosen required for
  // excising the signal
  //
  int_4 sample_freq = 8000 ;
  int_4 num_chan = 1;
  int_4 num_bytes = 2 ;
  
  sig_det.get_params_cc(sample_freq,num_chan,num_bytes);

  int_4 cur_time = (int_4)(endpts[0][0] * sample_freq);
  int_4 end_time = (int_4)(endpts[1][0] * sample_freq);
  int_4 num_samples = end_time - cur_time;
  int_4 samples_write;
  
  char_1* data_buffer = new char_1[(num_samples*num_chan*num_bytes)];
  
  // seek to the point in the file where speech begins.
  //
  fseek(fp_in, (cur_time*num_bytes*num_chan), SEEK_SET);
  
  // main read-write
  //
  while (cur_time < end_time) {
    
    // read data
    //
    int_4 samples_read = fread(data_buffer, num_chan*num_bytes,
			       num_samples, fp_in);
    
    if (samples_read < num_samples){
      samples_read = num_samples;
    }
    
    if((cur_time + samples_read) > (end_time))
      {samples_read = (int_4)(end_time - cur_time) ;}
    
    // output the data
    //
    if (samples_read != 0) {
      samples_write = fwrite(data_buffer, num_chan*num_bytes,
				 samples_read, fp_out);
    }
    else {
      return ISIP_TRUE;
    }
    
    // update counters
    //
    cur_time += samples_read;
  }

  //---------------------------------------------------------------------------
  //
  // exit
  //
  //---------------------------------------------------------------------------
  
  // close files
  //
  fclose(fp_in);
  fclose(fp_out);

  // exit gracefully
  //
  return(ISIP_NO_ERROR);
}



